/*
xmp2exif: restoring metadata in exif format from xmp data.
Copyright (C) 2009  Bram de Greve <bram.degreve@bramz.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <iostream>
#include <cassert>
#include <exiv2/image.hpp>
#include <exiv2/xmp.hpp>
#include <exiv2/exif.hpp>

template <typename accessor_type>
void xmp2exifProperty(
		const Exiv2::XmpData& xmp, Exiv2::ExifData& exif, 
		 const std::string& xmpKey, const std::string& exifKey,
		 accessor_type accessor)
{
	Exiv2::XmpData::const_iterator x = xmp.findKey(Exiv2::XmpKey(xmpKey));
	if (x != xmp.end())
	{
		//std::cout << xmpKey << ": '" << x->toString() << "'\n";
		Exiv2::ExifData::iterator y = exif.findKey(Exiv2::ExifKey(exifKey));
		if (y != exif.end())
		{
			exif.erase(y);
		}
		exif[exifKey] = accessor(*x);
	}
}



std::string asString(const Exiv2::Xmpdatum& x) 
{ 
	return x.toString(); 
}

Exiv2::Rational asRational(const Exiv2::Xmpdatum& x) 
{ 
	return x.toRational(); 
}

long asLong(const Exiv2::Xmpdatum& x) 
{ 
	return x.toLong(); 
}

std::string asDate(const Exiv2::Xmpdatum& x) 
{ 
	std::string date = x.toString().substr(0, 19);
	date[4] = ':';
	date[7] = ':';
	date[10] = ' ';
	return date;
}

std::string asLangAlt(const Exiv2::Xmpdatum& x)
{
	const std::string start = "lang=\"x-default\" ";
	std::string s = x.toString();
	if (s.find(start) == 0)
	{
		s = s.substr(start.length());
	}
	return s;
}

class Xmp2ExifCopier
{
public:
	Xmp2ExifCopier(const Exiv2::XmpData& xmp, Exiv2::ExifData& exif, const std::string& xmpSpace, const std::string& exifSpace):
		xmp_(xmp),
		exif_(exif),
		xmpSpace_("Xmp." + xmpSpace + "."),
		exifSpace_("Exif." + exifSpace + ".")
	{
	}
	const Xmp2ExifCopier& copyString(const std::string& xmpKey, const std::string& exifKey = "") const
	{
		return copy(xmpKey, exifKey, asString);
	}
	const Xmp2ExifCopier& copyLong(const std::string& xmpKey, const std::string& exifKey = "") const
	{
		return copy(xmpKey, exifKey, asLong);
	}
	const Xmp2ExifCopier& copyRational(const std::string& xmpKey, const std::string& exifKey = "") const
	{
		return copy(xmpKey, exifKey, asRational);
	}
	const Xmp2ExifCopier& copyDate(const std::string& xmpKey, const std::string& exifKey = "") const
	{
		return copy(xmpKey, exifKey, asDate);
	}
	const Xmp2ExifCopier& copyLangAlt(const std::string& xmpKey, const std::string& exifKey = "") const
	{
		return copy(xmpKey, exifKey, asLangAlt);
	}
private:
	template <typename Op> 
	const Xmp2ExifCopier& copy(const std::string& xmpKey, const std::string& exifKey, Op op) const
	{
		xmp2exifProperty(
			xmp_, exif_, 
			xmpSpace_ + xmpKey, exifSpace_ + (exifKey.empty() ? xmpKey : exifKey), 
			op);
		return *this;
	}
	const Exiv2::XmpData& xmp_;
	Exiv2::ExifData& exif_;
	std::string xmpSpace_;
	std::string exifSpace_;
};

void xmp2exif(const Exiv2::XmpData& xmp, Exiv2::ExifData& exif)
{
	Xmp2ExifCopier(xmp, exif, "exif", "Photo")
		.copyString("ExifVersion")
		.copyString("FlashpixVersion")
		.copyLong("ColorSpace")
		.copyRational("CompressedBitsPerPixel")
		.copyLong("PixelXDimension")
		.copyLong("PixelYDimension")
		.copyDate("DateTimeOriginal")
		.copyDate("DateTimeDigitized")
		.copyRational("ExposureTime")
		.copyRational("FNumber")
		.copyLong("ExposureProgram")
		.copyString("SpectralSensitivity")
		.copyLong("ISOSpeedRatings")
		.copyRational("ShutterSpeedValue")
		.copyRational("ApertureValue")
		.copyRational("BrightnessValue")
		.copyRational("ExposureBiasValue")
		.copyRational("MaxApertureValue")
		.copyRational("SubjectDistance")
		.copyLong("MeteringMode")
		.copyLong("LightSource")
		.copyRational("FocalLength")
		.copyRational("FlashEnergy")
		.copyRational("FocalPlaneXResolution")
		.copyRational("FocalPlaneYResolution")
		.copyLong("FocalPlaneResolutionUnit")
		.copyRational("ExposureIndex")
		.copyLong("SensingMethod")
		.copyLong("FileSource")
		.copyLong("SceneType")
		.copyLong("CustomRendered")
		.copyLong("ExposureMode")
		.copyLong("WhiteBalance")
		.copyLong("DigitalZoomRatio")
		.copyLong("FocalLengthIn35mmFilm")
		.copyLong("SceneCaptureType")
		.copyLong("GainControl")
		.copyLong("Contrast")
		.copyLong("Saturation")
		.copyLong("Sharpness")
		.copyLong("SubjectDistanceRange")
		.copyLong("ImageUniqueID")
		;
	Xmp2ExifCopier(xmp, exif, "tiff", "Image")
		.copyLong("ImageWidth")
		.copyLong("ImageLength")
		.copyLong("Compression")
		.copyLong("PhotometricInterpretation")
		.copyLong("Orientation")
		.copyLong("SamplesPerPixel")
		.copyLong("PlanarConfiguration")
		.copyLong("YCbCrPositioning")
		.copyRational("XResolution")
		.copyRational("YResolution")
		.copyRational("ResolutionUnit")
		.copyString("Make")
		.copyString("Model")
		;
	Xmp2ExifCopier(xmp, exif, "dc", "Image")
		.copyLangAlt("description", "ImageDescription")
		.copyLangAlt("creator", "Artist")
		.copyLangAlt("rights", "Copyright")
		;
	Xmp2ExifCopier(xmp, exif, "xmp", "Image")
		.copyLangAlt("ModifyDate", "DateTime")
		.copyLangAlt("CreatorTool", "Software")
		;
}

void xmp2exif(Exiv2::Image& image)
{
	image.readMetadata();
	const Exiv2::XmpData& xmp = image.xmpData();
	Exiv2::ExifData& exif = image.exifData();
	xmp2exif(xmp, exif);
	image.writeMetadata();
}

int main(int argc, char* argv[])
try
{
	if (argc < 2)
	{
		std::cout << "Usage: " << argv[0] << " file ..." << std::endl;
		return 1;
	}
	for (int i = 1; i < argc; ++i)
	{
		std::cout << argv[i] << std::endl;
		Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open(argv[i]);
		assert(image.get());
		xmp2exif(*image);
	}
	return 0;
}
catch (const std::exception& error) 
{
	std::cerr << "Error: " << error.what();
    return 2;
}
